package jeliot.lang;

import java.util.*;
import jeliot.parser.*;
import jeliot.parser.error.*;

/**
  * @author Pekka Uronen
  *
  * created         27.9.1999
  */
public class ArrayType extends ReferenceType {
    
    private static Hashtable instanceIndex = new Hashtable();
    
    private Type componentType;
    
    private FieldPointer lengthPointer = new ArrayLengthPointer(this);
    
    private ArrayType() { }
    
    private ArrayType(Type componentType) {
        this.componentType = componentType;
    }
    
    public static ArrayType getType(Class c) {
        Class compo = c.getComponentType();
        return getType(Type.getTypeByClass(compo), 1);
    }
    
    public static ArrayType getType(Type componentType, int dim) {
        if (dim < 1) {
            return null;
        }
        if (dim > 1) {
            componentType = getType(componentType, dim-1);
        }
        ArrayType at = (ArrayType)instanceIndex.get(componentType); 
        if (at == null) {
            at = new ArrayType(componentType);
            instanceIndex.put(componentType, at); 
        }
        return at;
    }
    
    public Type getComponentType() {
        return componentType;
    }
    
    /** Checks if a value of this type may be assigned to a variable of
      * given type t.
      *
      * see [GJS96, chapt. 5.2]
      */
    public boolean mayBeAssignedTo(Type t) {
        // array may always be assigned to an object variable.
        if (t == ReferenceType.ObjectTYPE) {
            return true;
        }
        // array may be assigned to a cloneable variable.
        if (t == ReferenceType.CloneableTYPE) {
            return true;    
        }
        // t must be an array
        if (! (t instanceof ArrayType) ) {
            return false;
        }
        ArrayType at = (ArrayType)t;
        
        // If the component type of either array is primitive, then
        // both must be of the same primitive type.
        if (componentType instanceof PrimitiveType ||
            at.componentType instanceof PrimitiveType) {
            
            return componentType == at.componentType;
        }
        // Both component types are reference types. Recursively check
        // the assignment of the components.
        return componentType.mayBeAssignedTo(at.componentType);
    }
    
    /** Checks if a value of this type can be cast to a value of type
      * t in an explicit cast. If 
      */
    public boolean mayBeCastTo(Type t) {
        // array may be casted to object.
        if (t == ReferenceType.ObjectTYPE) {
            return true;
        }
        // array may be casted to cloneable.
        if (t == ReferenceType.CloneableTYPE) {
            return true;    
        }
        // t must be an array
        if (! (t instanceof ArrayType) ) {
            return false;
        }
        ArrayType at = (ArrayType)t;
        
        // If the component type of either array is primitive, then
        // both must be of the same primitive type.
        if (componentType instanceof PrimitiveType ||
            at.componentType instanceof PrimitiveType) {
            
            return componentType == at.componentType;
        }
        // Both component types are reference types. Recursively check
        // the cast of the components.
        return componentType.mayBeCastTo(at.componentType);
    }
    
    public MethodPointer getMethod(String name, Type[] argt) {
        return null;
    }
    
    public FieldPointer getField(String name) {
        if (! name.equals("length")) {
            throw new NonexistentFieldException(this, name);
        }
        return lengthPointer;    
    }
    
    public String toString() {
        return componentType.toString() + "[]";
    }
    
}