package jeliot.lang;

import java.util.*;
import jeliot.parser.*;

/**
  * @author Pekka Uronen
  *
  * created       27.9.1999
  */
public class PrimitiveType extends Type {

    private static final int BOOLEAN   = 0;
    private static final int BYTE      = 1;
    private static final int SHORT     = 2;
    private static final int INT       = 3;
    private static final int LONG      = 4;
    private static final int CHAR      = 5;
    private static final int FLOAT     = 6;
    private static final int DOUBLE    = 7;

    private static final Class[] primitiveClass =
      { boolean.class, byte.class, short.class,
        int.class, long.class, char.class, float.class,
        double.class
      };
      
    private static final Object[] neutral =
      { Boolean.FALSE, new Byte((byte)0), new Short((short)0), 
        new Integer(0), new Long(0l), new Character('\0'), 
        new Float(0.0f), new Double(0.0)
      };      

    public static final PrimitiveType booleanTYPE
            = new PrimitiveType(BOOLEAN);
    public static final PrimitiveType byteTYPE
            = new PrimitiveType(BYTE);
    public static final PrimitiveType shortTYPE
            = new PrimitiveType(SHORT);
    public static final PrimitiveType intTYPE
            = new PrimitiveType(INT);
    public static final PrimitiveType longTYPE
            = new PrimitiveType(LONG);
    public static final PrimitiveType charTYPE
            = new PrimitiveType(CHAR);
    public static final PrimitiveType floatTYPE
            = new PrimitiveType(FLOAT);
    public static final PrimitiveType doubleTYPE
             = new PrimitiveType(DOUBLE);
             
    private static PrimitiveType[] typeIndex = {
        booleanTYPE, byteTYPE, shortTYPE, intTYPE, longTYPE,
        charTYPE, floatTYPE, doubleTYPE
    };
             
    private static PrimitiveType[] unaryPromotion = {
        null, 
        intTYPE, intTYPE, intTYPE, longTYPE, intTYPE,
        floatTYPE, doubleTYPE
    };
    
             
    private static PrimitiveType[][] binaryPromotion = {
        {   null, null, null, null, null, null, null},
        {   null, intTYPE, intTYPE, intTYPE, 
                 longTYPE, intTYPE, floatTYPE, doubleTYPE},
        {   null, intTYPE, intTYPE, intTYPE, 
                 longTYPE, intTYPE, floatTYPE, doubleTYPE},
        {   null, intTYPE, intTYPE, intTYPE, 
            longTYPE, intTYPE, floatTYPE, doubleTYPE},
        {   null, longTYPE, longTYPE, longTYPE, 
            longTYPE, longTYPE, floatTYPE, doubleTYPE},
        {   null, intTYPE, intTYPE, intTYPE, 
            longTYPE, intTYPE, floatTYPE, doubleTYPE},
        {   null, floatTYPE, floatTYPE, floatTYPE, 
            floatTYPE, floatTYPE, floatTYPE, doubleTYPE},
        {   null, doubleTYPE, doubleTYPE, doubleTYPE, 
            doubleTYPE, doubleTYPE, doubleTYPE, doubleTYPE}
    };
            
    // assignable[int][char] means: int may be assigned to char.
    private static boolean assignable[][] = {
    // boolean   byte  short    int   long   char  float double
        { true, false, false, false, false, false, false, false},
        {false,  true,  true,  true,  true,  true,  true,  true},
        {false, false,  true,  true,  true,  true,  true,  true},
        {false, false, false,  true,  true, false,  true,  true},
        {false, false, false, false,  true, false,  true,  true},
        {false, false,  true,  true,  true,  true,  true,  true},
        {false, false, false, false, false, false,  true,  true},
        {false, false, false, false, false, false, false,  true}
    };

    private int index;

    /** Instantiating primitive types is prohibited for other classes.
      */
    private PrimitiveType() { }

    private PrimitiveType(int index) {
        this.index = index;
    }
    
    /** Returns the primitive type corresponding the given class.
      * If given class does not represent a primitive type, this 
      * method returns null.
      */
    public static PrimitiveType getType(Class c) {
        for (int i = 0; i < primitiveClass.length; ++i) {
            if (c == primitiveClass[i]) {
                return typeIndex[i];
            }
        }
        return null;
    }
    

    /** Checks if a value of this type can be assigned to a variable
      * of type t. 
      */
    public boolean mayBeAssignedTo(Type t) {
        if (!(t instanceof PrimitiveType)) {
            return false;
        }
        // t is a primitive type
        PrimitiveType pt = (PrimitiveType)t;
        return assignable[index][pt.index];
    }
    
    /** Checks if a value of this type can be cast to a value of type
      * t in an explicit cast. If 
      */
    public boolean mayBeCastTo(Type t) {
        if (!(t instanceof PrimitiveType)) {
            return false;
        }
        // t is a primitive type
        return (this == booleanTYPE) == (t == booleanTYPE);
    }
    
    /** Checks if this type is a numeric type. Only primitive types
      * other than boolean return true.
      */
    public boolean isNumeric() {
        return this != booleanTYPE;
    }
    
    /** Checks if this type is an integral type.
      */
    public boolean isIntegral() {
        return BOOLEAN < index && index < FLOAT;
    }
    
    /** Returns the unary numeric promoted type of this type.
      */ 
    public PrimitiveType unaryNumericPromotion() {
        return unaryPromotion[index];
    }
    
    /** Returns the binary numeric promoted type of arguments types.
      */
    public static PrimitiveType binaryNumericPromotion (
            PrimitiveType t1, PrimitiveType t2) {
                
        return binaryPromotion[t1.index][t2.index];
    }
    
    /** Returns the index of this type. This index may be used to 
      * create arrays based on primitive types.
      */
     public int getIndex() {
        return index;
    }
    
    public String toString() {
        return primitiveClass[index].getName();
    }
    
    public Value neutralValue() {
        return new Value(this, neutral[index]);
    }
    
}