package jeliot.parser;

import java.util.*;
import jeliot.lang.*;

/**
  * @author Pekka Uronen
  *
  * created         24.8.1999
  */
public class NameSpace extends ProgramTreeVisitor {

    /** The name of the package compilation unit under
      * examination belongs to.
      */ 
    private String thisPackage;

    /** Contains the names of the packages imported-on-demand.
      */
    private Vector packages = new Vector();

    /** Contains the single-imported classes as ForeignType instances,
      * indexed with their simple names.
      */
    private Hashtable imported = new Hashtable();

    /** Contains the classes and interfaces declared in this
      * compilation unit as PClass objects, indexed with their
      * simple names.
      */
    private Hashtable declared = new Hashtable();

    /** Contains the classes found by searching the packages 
      * imported-on-demand, indexed with their simple names.
      * This is used as a shortcut to find often used classes.
      * Classes are represented by ForeignType instances.
      */
    private Hashtable onDemand = new Hashtable();
      
    /** Current class path. This is used to verify existence of
      * imported packages.
      */
    private String classPath =
            System.getProperties().getProperty("java.class.path");


    public NameSpace(PCompilationUnit unit) {
        unit.acceptVisitor(this);
    }

    /** Collects the import declarations in given compilation unit.
      */
    public void visitCompilationUnit(PCompilationUnit cu) {
        this.thisPackage = cu.getPackageName();
        
        // java.lang is imported in every compilation unit.
        importPackage("java.lang");

        // search imports and import.
        Enumeration enum = cu.getImports();
        if (enum != null) {
            while (enum.hasMoreElements()) {
                PImport impo = (PImport)enum.nextElement();
                if (impo.isOnDemand()) {
                    importPackage(impo.getPackageName());
                }
                else {
                    importClass(impo.getName().toString());
                }
            }
        }
    }

    /** Adds the visited class to declarations.
      */
    public void visitClass(PClass c) {
        declareClass(c);
    }

    /** Imports the package with given name to this name space.
      * The name must be in form <CODE>java.lang</CODE>.
      */
    void importPackage(String packageName) {
        // if the package is allready imported, do nothing.
        if (!packages.contains(packageName)) {

            // here should be a check for package validity.
            
            packages.addElement(packageName);
        }   
    }

    /** Single-type-inports a class or interface.
      * There is no check for syntactic correctness of the name, as
      * the syntax is assumed to be checked by the parser.
      *
      * @param qname Qualified name of the class to be imported.
      */
    void importClass(String qname) {
        // Extract the simple name.
        int doti = qname.lastIndexOf('.');
        String sname = (doti == -1) ?
                qname   :
                qname.substring(doti + 1);

        if (imported.get(sname) != null) {
            throw new SemanticException("Class " + qname +
                    " imported twice.");
        }
        try {    
            Class c = Class.forName(qname);
            imported.put(sname, ForeignType.getType(c));
        }
        catch (ClassNotFoundException e) {

            // here should be a check for package validity.
            
            throw new SemanticException("Class " + qname +
                    " does not exist.");
        }
    }

    /** Notifies of a class or interface declared in current
      * compilation unit.
      *
      * @param c The PClass object of the class to be declared.
      */
    void declareClass(PClass c) {
        String sname = c.getName();
        if (imported.get(sname) != null) {
            throw new SemanticException("Class " + sname +
                    " both imported and declared.");
        }
        if (declared.get(sname) != null) {
            throw new SemanticException("Class " + sname +
                    " declared twice.");
        }
        declared.put(sname, c);       
    }

    /** Finds a class for given simple or qualified name.
      * This algorithm is described in [GJS96:6.5.4].
      *
      * @returns A ReferenceType of the matching class, or
      *     <CODE>null</CODE>, if none was found.
      */
    ReferenceType findClass(String name) {
        int doti = name.lastIndexOf('.');
        if (doti == -1) {
            // Simple name
            
            // search single-type-imports
            ReferenceType type = (ForeignType)imported.get(name);
            if (type != null) {
                return type; 
            }      
            // search current compilation unit
            PClass dc = (PClass)declared.get(name);
            if (dc != null) {
                return  dc.getType();
            }
            // Frequently used on-demand classes
            type = (ForeignType)onDemand.get(name);
            if (type != null) {
                return type;
            }
            // other compilation units of this package
            if (thisPackage != null) {
                try {
                    // if the class is not in the package,
                    // next statement will throw a
                    // ClassNotFoundException.
                    Class c = Class.forName(thisPackage+"."+name);
                    type = ForeignType.getType(c);
                    onDemand.put(name, type);
                    return type;
                }
                catch (ClassNotFoundException e) { }
            }
            // import-on-demands
            String foundPack = null;
            Enumeration enum = packages.elements();
            while (enum.hasMoreElements()) {
                String pack = (String)enum.nextElement();
                try {
                    // if the class is not in the package,
                    // next statement will throw a
                    // ClassNotFoundException.
                    Class c = Class.forName(pack+"."+name);
        
                    if (type == null) {
                        type = ForeignType.getType(c);
                        foundPack = pack;
                        onDemand.put(name, type);
                    }
                    else {
                        throw new SemanticException(
                            "Type "+name+" declared both" +
                            " in package " + foundPack + 
                            " and package " + pack + ".");
                    }
                }
                catch (ClassNotFoundException e) { }
            }
            return type;
        }
        else {
            // Qualified name
            
            // User defined classes cannot be accessed this way -
            // they reside in the default unnamed package. So it is
            // enough to find a foreign (compiled) class.

            try {
                Class c = Class.forName(name);
                ForeignType type = ForeignType.getType(c);
                return type;
            }
            catch (ClassNotFoundException e) {
                return null;
            }
        }
    }

}
