package jeliot.parser;

import jeliot.lang.*;
import jeliot.theatre.*;
import jeliot.parser.error.*;
import jeliot.FeatureNotImplementedException;

/**
  * PUnaryExpr represents a binary expression (*, +, ==, etc.).
  *
  * @author Pekka Uronen
  *
  * created         6.8.1999
  */
public class PBinaryExpr extends PExpression {

    public static final int MULT        = 0;
    public static final int DIV         = 1;
    public static final int MOD         = 2;
    public static final int PLUS        = 3;
    public static final int MINUS       = 4;
    public static final int LSHIFT      = 5;
    public static final int RSHIFT      = 6;   
    public static final int URSHIFT     = 7;
    public static final int LT          = 8;
    public static final int GT          = 9;
    public static final int LTEQ        = 10;
    public static final int GTEQ        = 11;
    public static final int INSTANCEOF  = 12;
    public static final int EQEQ        = 13;
    public static final int NOTEQ       = 14;
    public static final int AND         = 15;
    public static final int XOR         = 16;
    public static final int OR          = 17;
    public static final int ANDAND      = 18;
    public static final int OROR        = 19;
    
    private int operator;
    
    private PExpression expr1;
    private PExpression expr2;
    
    public PBinaryExpr(
            PExpression expr1,
            PExpression expr2,
            int operator) {
                
        this.expr1 = expr1;
        this.expr2 = expr2;
        this.operator = operator;

        expr1.setParent(this);
        expr2.setParent(this);
    }

    public void acceptVisitor(ProgramTreeVisitor visitor) {    
        visitor.enterBinaryExpression(this);
        expr1.acceptVisitor(visitor);
        expr2.acceptVisitor(visitor);
        visitor.leaveBinaryExpression(this);
    }

    public int getOperator() {
        return operator;
    }
    
    public int getLeftPos() {
        return expr1.getLeftPos();
    }
    
    public int getRightPos() {
        return expr2.getRightPos();
    }
    
    public PExpression getLeftHand() {
        return expr1;
    }

    public PExpression getRightHand() {
        return expr2;
    }

    public Value evaluate(Director director) {
        super.evaluate(director);
        director.atExpression(this);

        // Check if expression should be animated in two parts.
        // if twoPart is true, the left hand value is taken to the
        // expression area before evaluating the right hand expression.
        boolean twoPart =
                (
                    !expr1.usesScratch() && !expr2.isPrimary()
                ) ||
                    operator == ANDAND || operator == OROR;

        // Evaluate the left hand expression
        Value lval = expr1.evaluate(director);
        Object lv = lval.getValue();

        // If two part animation, show the first part.
        ExpressionActor expact = null;
        if (twoPart) {
            expact = director.beginBinaryExpression(lval, operator);
        }

        // If this is not a conditional expression, evaluate the
        // right hand expression.
        Value rval;
        Object rv;
        switch (operator) {
            case (ANDAND):
            case (OROR):
                rval = null;
                rv = null;
                break;
            default:
                rval = expr2.evaluate(director);
                rv = rval.getValue();
        }

        // Evaluate this expression, depending on the type of the
        // operator.
        Value result;
        switch (operator) {
            case (EQEQ): 
                result = Value.newBoolean(areEqual(lv, rv));
                break;
                
            case (NOTEQ): 
                result = Value.newBoolean(!areEqual(lv, rv));
                break;

            case (LT): {
                Type t = PrimitiveType.binaryNumericPromotion(
                        (PrimitiveType)lval.getType(), 
                        (PrimitiveType)rval.getType());
                if (t == PrimitiveType.intTYPE || 
                        t == PrimitiveType.longTYPE) {
                            
                    result = Value.newBoolean(
                        ( (Number)lv ).longValue() <
                        ( (Number)rv ).longValue() );
                }
                else {
                    result = Value.newBoolean(
                        ( (Number)lv ).doubleValue() <
                        ( (Number)rv ).doubleValue() );
                }   
                break;
            }
            case (GT): {
                Type t = PrimitiveType.binaryNumericPromotion(
                        (PrimitiveType)lval.getType(), 
                        (PrimitiveType)rval.getType());
                if (t == PrimitiveType.intTYPE || 
                        t == PrimitiveType.longTYPE) {
                            
                    result = Value.newBoolean(
                        ( (Number)lv ).longValue() >
                        ( (Number)rv ).longValue() );
                }
                else {
                    result = Value.newBoolean(
                        ( (Number)lv ).doubleValue() >
                        ( (Number)rv ).doubleValue() );
                }   
                break;
            }
            case (LTEQ): {
                Type t = PrimitiveType.binaryNumericPromotion(
                        (PrimitiveType)lval.getType(), 
                        (PrimitiveType)rval.getType());
                if (t == PrimitiveType.intTYPE || 
                        t == PrimitiveType.longTYPE) {
                    
                    result = Value.newBoolean(
                        ( (Number)lv ).longValue() <=
                        ( (Number)rv ).longValue() );
                }
                else {
                    result = Value.newBoolean(
                        ( (Number)lv ).doubleValue() <=
                        ( (Number)rv ).doubleValue() );
                }   
                break;
            }
            case (GTEQ): {
                Type t = PrimitiveType.binaryNumericPromotion(
                        (PrimitiveType)lval.getType(), 
                        (PrimitiveType)rval.getType());
                if (t == PrimitiveType.intTYPE || 
                        t == PrimitiveType.longTYPE) {
                        
                    result = Value.newBoolean(
                        ( (Number)lv ).longValue() >=
                        ( (Number)rv ).longValue() );
                }
                else {
                    result = Value.newBoolean(
                        ( (Number)lv ).doubleValue() >=
                        ( (Number)rv ).doubleValue() );
                }   
                break;
            }
            case (PLUS): 
                result = new Value(getType(), evalPlus(lv, rv));
                break;

            case (MINUS):
                result = new Value(getType(), evalMinus(lv, rv));
                break;

            case (MULT):
                result = new Value(getType(), evalMult(lv, rv));
                break;

            case (DIV):
                try {
                    result = new Value(getType(), evalDiv(lv, rv));
                }
                catch (ArithmeticException e) {
                    result = new Value(getType(), e);
                }
                break;
                
            case (MOD):
                try {
                    result = new Value(getType(), evalMod(lv, rv));
                }
                catch (ArithmeticException e) {
                    result = new Value(getType(), e);
                }
                break;

            case (ANDAND): {
                boolean b = ((Boolean)lv).booleanValue();
                if (b) {
                    rval = expr2.evaluate(director);
                    rv = rval.getValue();
                    boolean br = ((Boolean)rv).booleanValue();
                    result = new Value(PrimitiveType.booleanTYPE,
                            new Boolean(br));
                }
                else {
                    // If the left hand value is false, there is no
                    // need to evaluate the right hand expression.
                    rval = null;
                    result = new Value(PrimitiveType.booleanTYPE,
                            new Boolean(false));
                }
                break;
            }
            case (OROR): {
                boolean b = ((Boolean)lv).booleanValue();
                if (b) {
                    // If the left hand value is true, there is no
                    // need to evaluate the right hand expression.
                    rval = null;
                    result = new Value(PrimitiveType.booleanTYPE,
                            new Boolean(true));
                }
                else {
                    rval = expr2.evaluate(director);
                    rv = rval.getValue();
                    boolean br = ((Boolean)rv).booleanValue();
                    result = new Value(PrimitiveType.booleanTYPE,
                            new Boolean(br));
                }
                break;
            }
            default:
                throw new RuntimeException("Strange operator: "+operator);
        }
        director.atExpression(this);
        if (!twoPart) {
            director.animateBinaryExpression(operator, lval, rval, result);
        }
        else {
            director.finishBinaryExpression(rval, result, operator, expact);
        }
        return result;
        
    }

    private boolean areEqual(Object lv, Object rv) {
        if (    lv instanceof Number || lv instanceof Character &&
                rv instanceof Number || rv instanceof Character) {

            if (lv instanceof Double || lv instanceof Float) {
                // left one is floating point
                return  ( (Number)lv ).doubleValue() ==
                        (   (rv instanceof Character)       ?
                            ( (Character)rv ).charValue()   :
                            ( (Number)rv ).doubleValue()    );
            }
            else if (rv instanceof Double || rv instanceof Float) {
                // right hand side is floating point
                return  ( (Number)rv ).doubleValue() ==
                        (   (lv instanceof Character)       ?
                            ( (Character)lv ).charValue()   :
                            ( (Number)lv ).doubleValue()    );
            }
            else {
                // neither one is floating point
                return (    (lv instanceof Number)          ?
                            ( (Number)lv ).longValue()      :
                            ( (Character)lv ).charValue()   ) ==
                       (    (rv instanceof Number)          ?
                            ( (Number)rv ).longValue()      :
                            ( (Character)rv ).charValue()   );
            }
        }
        else {
            return lv.equals(rv);
        }
    }
    
    private Object evalPlus(Object lv, Object rv) {
        Type t = getType();

        if (lv instanceof Character) {
            lv = new Integer( ((Character)lv).charValue() );
        }
        if (rv instanceof Character) {
            rv = new Integer( ((Character)rv).charValue() );
        }
        if (t == PrimitiveType.intTYPE) {
            int l = ( (Number)lv ).intValue();
            int r = ( (Number)rv ).intValue();
            return new Integer(l + r);
        }
        else if (t == PrimitiveType.longTYPE) {
            long l = ( (Number)lv ).longValue();
            long r = ( (Number)rv ).longValue();
            return new Long(l + r);
        }
        else if (t == PrimitiveType.floatTYPE) {
            float l = ( (Number)lv ).floatValue();
            float r = ( (Number)rv ).floatValue();
            return new Float(l + r);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            double l = ( (Number)lv ).doubleValue();
            double r = ( (Number)rv ).doubleValue();
            return new Double(l + r);
        }
        return null;            
    }

    private Number evalMinus(Object lv, Object rv) {
        Type t = getType();
        if (lv instanceof Character) {
            lv = new Integer( ((Character)lv).charValue() );
        }
        if (rv instanceof Character) {
            rv = new Integer( ((Character)rv).charValue() );
        }
        if (t == PrimitiveType.intTYPE) {
            int l = ( (Number)lv ).intValue();
            int r = ( (Number)rv ).intValue();
            return new Integer(l - r);
        }
        else if (t == PrimitiveType.longTYPE) {
            long l = ( (Number)lv ).longValue();
            long r = ( (Number)rv ).longValue();
            return new Long(l - r);
        }
        else if (t == PrimitiveType.floatTYPE) {
            float l = ( (Number)lv ).floatValue();
            float r = ( (Number)rv ).floatValue();
            return new Float(l - r);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            double l = ( (Number)lv ).doubleValue();
            double r = ( (Number)rv ).doubleValue();
            return new Double(l - r);
        }
        return null;                
    }

    private Number evalMult(Object lv, Object rv) {
        Type t = getType();
        if (lv instanceof Character) {
            lv = new Integer( ((Character)lv).charValue() );
        }
        if (rv instanceof Character) {
            rv = new Integer( ((Character)rv).charValue() );
        }
        if (t == PrimitiveType.intTYPE) {
            int l = ( (Number)lv ).intValue();
            int r = ( (Number)rv ).intValue();
            return new Integer(l * r);
        }
        else if (t == PrimitiveType.longTYPE) {
            long l = ( (Number)lv ).longValue();
            long r = ( (Number)rv ).longValue();
            return new Long(l * r);
        }
        else if (t == PrimitiveType.floatTYPE) {
            float l = ( (Number)lv ).floatValue();
            float r = ( (Number)rv ).floatValue();
            return new Float(l * r);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            double l = ( (Number)lv ).doubleValue();
            double r = ( (Number)rv ).doubleValue();
            return new Double(l * r);
        }
        return null;                
    }

    private Number evalDiv(Object lv, Object rv) {
        Type t = getType();
        if (lv instanceof Character) {
            lv = new Integer( ((Character)lv).charValue() );
        }
        if (rv instanceof Character) {
            rv = new Integer( ((Character)rv).charValue() );
        }
        if (t == PrimitiveType.intTYPE) {
            int l = ( (Number)lv ).intValue();
            int r = ( (Number)rv ).intValue();
            return new Integer(l / r);
        }
        else if (t == PrimitiveType.longTYPE) {
            long l = ( (Number)lv ).longValue();
            long r = ( (Number)rv ).longValue();
            return new Long(l / r);
        }
        else if (t == PrimitiveType.floatTYPE) {
            float l = ( (Number)lv ).floatValue();
            float r = ( (Number)rv ).floatValue();
            return new Float(l / r);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            double l = ( (Number)lv ).doubleValue();
            double r = ( (Number)rv ).doubleValue();
            return new Double(l / r);
        }
        return null;                
    }

    private Number evalMod(Object lv, Object rv) {
        Type t = getType();
        if (lv instanceof Character) {
            lv = new Integer( ((Character)lv).charValue() );
        }
        if (rv instanceof Character) {
            rv = new Integer( ((Character)rv).charValue() );
        }
        if (t == PrimitiveType.intTYPE) {
            int l = ( (Number)lv ).intValue();
            int r = ( (Number)rv ).intValue();
            return new Integer(l % r);
        }
        else if (t == PrimitiveType.longTYPE) {
            long l = ( (Number)lv ).longValue();
            long r = ( (Number)rv ).longValue();
            return new Long(l % r);
        }
        return null;                
    }    
    
    
    public void typeCheck(
            NameSpace space,
            LocalVariableBundle locals,
            PClass context) {
            
        Type type = getType();
        Type t1 = expr1.getType();
        Type t2 = expr2.getType();

        switch (operator) {
            case (PLUS):
                if (t1 == ReferenceType.StringTYPE) {
                    setType(t1);
                    break;
                }
                else if (t2 == ReferenceType.StringTYPE) {
                    setType(t2);
                    break;
                }
                // falls through
            case (MINUS):
            case (MULT):
            case (DIV):
            case (MOD):
                if ( !t1.isNumeric() || !t2.isNumeric() ) {
                    throw new NoGoodArgumentException(this);
                }
                setType(PrimitiveType.binaryNumericPromotion(
                        (PrimitiveType)t1, (PrimitiveType)t2));
                break;

            case (RSHIFT):
            case (LSHIFT):
            case (URSHIFT):
                if ( !t1.isIntegral() || !t2.isIntegral() ) {
                    throw new NoGoodArgumentException(this);
                }
                setType(((PrimitiveType)t1).unaryNumericPromotion());
                break;

            case (LT):
            case (GT):
            case (LTEQ):
            case (GTEQ):
                if ( !t1.isNumeric() || !t2.isNumeric() ) {
                    throw new NoGoodArgumentException(this);
                }
                setType(PrimitiveType.booleanTYPE);
                break;

            case (INSTANCEOF):
                throw new FeatureNotImplementedException("Instanceof");

            case (EQEQ):
            case (NOTEQ):
                if (!t1.mayBeCastTo(t2) && !t2.mayBeCastTo(t1)) {
                    throw new NoGoodArgumentException(this);
                }
                setType(PrimitiveType.booleanTYPE);
                break;

            case (AND):
            case (XOR):
            case (OR):
                if (t1.isIntegral() && t2.isIntegral()) {
                    setType(((PrimitiveType)t1).
                            unaryNumericPromotion());
                }
                else if (   t1 != PrimitiveType.booleanTYPE ||
                            t2 != PrimitiveType.booleanTYPE   ) {

                    throw new NoGoodArgumentException(this);
                }
                setType(PrimitiveType.booleanTYPE);
                break;

            case (ANDAND):
            case (OROR):
                if (   t1 == PrimitiveType.booleanTYPE &&
                       t2 == PrimitiveType.booleanTYPE   ) {

                    setType(PrimitiveType.booleanTYPE);
                }
                else {
                    throw new NoGoodArgumentException(this     );
                }
                break;

            default:
                throw new RuntimeException();
        }
    }
    
}
